/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlTrainingDatasetBuilder.cpp
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**              based on applications by
**                     Stephen Gould <sgould@stanford.edu>
**                     Ian Goodfellow <ia3n@cs.stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <sys/types.h>
#include <limits>

#include "svlBase.h"
#include "svlVision.h"
#include "svlTrainingDatasetBuilder.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

using namespace std;

// related to all extractions
set<string> svlTrainingDatasetBuilder::OBJECTS;
bool svlTrainingDatasetBuilder::ALL_WINDOWS = false;

// related to negatives
bool svlTrainingDatasetBuilder::SKIP_NEG = false;
bool svlTrainingDatasetBuilder::FPOS_ONLY = false;
string svlTrainingDatasetBuilder::DETECTIONS_FILE = string("");
double svlTrainingDatasetBuilder::THRESHOLD = 0.0;
bool svlTrainingDatasetBuilder::INCLUDE_OTHER_OBJS = false;
int svlTrainingDatasetBuilder::NUM_NEG = 100;
int svlTrainingDatasetBuilder::NEG_HEIGHT = 32;
double svlTrainingDatasetBuilder::NEG_ASPECT_RATIO = 1.0;

// related to positives
bool svlTrainingDatasetBuilder::SKIP_POS = false;
bool svlTrainingDatasetBuilder::POS_BEST_WINDOW = false;

// related to writer
string svlTrainingDatasetBuilder::BASE_DIR = string(".");
bool svlTrainingDatasetBuilder::CREATE_DIRS = false;
string svlTrainingDatasetBuilder::NEG_SUBDIR_NAME = string("negative");
string svlTrainingDatasetBuilder::POS_PREFIX = string("");
string svlTrainingDatasetBuilder::NEG_PREFIX = string("");
bool svlTrainingDatasetBuilder::INCLUDE_FLIPPED = false;
int svlTrainingDatasetBuilder::RESIZE_WIDTH = -1;
int svlTrainingDatasetBuilder::RESIZE_HEIGHT = -1;
bool svlTrainingDatasetBuilder::USE_WIN_REFS = false;



svlTrainingDatasetBuilder::svlTrainingDatasetBuilder(const char *imageSeqFilename,
						     const char *gtFilename) :
  _bAllWindows(ALL_WINDOWS), _bSkipNeg(SKIP_NEG), _bFPonly(FPOS_ONLY),
  _threshold(THRESHOLD), _bIncludeOthers(INCLUDE_OTHER_OBJS),
  _numNeg(NUM_NEG), _negHeight(NEG_HEIGHT), _negAspectRatio(NEG_ASPECT_RATIO),
  _bSkipPos(SKIP_POS), _bPosBestWindow(POS_BEST_WINDOW)
{
  // loads the image filenames
  SVL_LOG(SVL_LOG_VERBOSE, "Reading image sequence from " << imageSeqFilename << "...");
  _imageSequence.load(imageSeqFilename);
  SVL_LOG(SVL_LOG_VERBOSE, "..." << _imageSequence.size() << " images read");

  if (!hasHomogeneousExtensions(_imageSequence)) {
    SVL_LOG(SVL_LOG_FATAL, "Sequence file " << imageSeqFilename
	    << " has multiple file extensions, so no standardized swapping "
	    << "of file extensions will work for managing channels.");
  }

  // load the object labels
  SVL_LOG(SVL_LOG_VERBOSE, "Reading object labels from " << gtFilename << "...");
  readObject2dSequence(gtFilename, _groundTruth);
  SVL_LOG(SVL_LOG_VERBOSE, "..." << _groundTruth.size() << " labeled frames read");

  // loads the detections, if any
  if (DETECTIONS_FILE.length() > 0) {
    SVL_LOG(SVL_LOG_VERBOSE, "Reading detections from " << DETECTIONS_FILE << "...");
    readObject2dSequence(DETECTIONS_FILE.c_str(), _objectDetections);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << _objectDetections.size() << " labeled frames read");
    int total = countObjects(_objectDetections);
    int removed = nonMaximalSuppression(_objectDetections, 0.25, 0.25, 0.64);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << removed << " out of " << total << " non-maximal objects suppressed");
  }  

  // create the directory for negative objects
  if (CREATE_DIRS && !_bSkipNeg) {
    string dirName = string(BASE_DIR) + string("/") + string(NEG_SUBDIR_NAME);
    svlCreateDirectory(dirName.c_str());
  }
  
  // loads the base extension for file reader from the image sequence
  svlImageLoader::BASE_EXT = getExtension(_imageSequence);

  // _writer by default initialized with relevant parameters
  _writer = new svlClipWriter();
  _loader = new svlImageLoader(false);
  _wExtractor = new svlImageWindowExtractor(OBJECTS, RESIZE_WIDTH, RESIZE_HEIGHT);
}

svlTrainingDatasetBuilder::~svlTrainingDatasetBuilder()
{
  delete _loader;
  delete _wExtractor;
  delete _writer;
}

void svlTrainingDatasetBuilder::writeDataset(int parallelCount, int parallelId)
{
  // iterate over the channels as the outer loop now
  for (unsigned chNum = 0; chNum < _loader->numChannels(); chNum++) {
    
    if (_loader->channelTypes(chNum) == SVL_CHANNEL_EDGE) {
      SVL_LOG(SVL_LOG_WARNING, "trainingDatasetBuilder does not export edge patches. "
	      << "Just make them on the fly in later apps.");
      continue;
    }

    svlPatchDefinitionType chPatchType = _loader->patchTypes(chNum);
    string chExt = _loader->extensions(chNum);

    _loader->printChannelInfo(chNum);

    // load images and snip regions
    for (unsigned i = 0; i < _imageSequence.size(); i++) {
      if (parallelCount > 0) {
        if ((i % unsigned(parallelCount)) != unsigned(parallelId)) continue ;
      }

      SVL_LOG(SVL_LOG_VERBOSE, "Processing image " << _imageSequence[i] << "...");

      string filename = _imageSequence.directory() + string("/") + _imageSequence[i];

      IplImage *image = _loader->getSimpleImageFromFilename(filename.c_str(), chNum);
      if (!image) 
	continue;
      
      //look up the positive examples for this frame
      string imageId = strBaseName(_imageSequence[i]);
      if (_groundTruth.find(imageId) != _groundTruth.end()) 
	_wExtractor->setGroundTruthLabels(_groundTruth[imageId]);
      else
	_wExtractor->clearGroundTruthLabels();

      _wExtractor->setImageSize(image->width, image->height);

      svlObject2dFrame negWindows;
      svlObject2dFrame posWindows;

      if (_bAllWindows) {
	// includes all windows

	// avoids having to fill in the vector with a potentially very
	// long list of negatives unless necessary
	if (_bSkipPos) {
	  _wExtractor->getAllNegWindows(negWindows);
	} else if (_bSkipNeg) {
	  _wExtractor->getAllPosWindows(posWindows);
	} else {
	  _wExtractor->getAllWindows(posWindows, negWindows);
	}
      }

      else {
	// not all windows should be included

	if (!_bSkipPos) {
	  // finds positive windows
	  if (_bPosBestWindow)
	    _wExtractor->getBestPositiveWindows(posWindows);
	  else
	    _wExtractor->getPositives(posWindows);
	}
	if (!_bSkipNeg) {
	  // finds negative windows
	  
	  // (1) false positive detections, if available
	  if (_objectDetections.find(imageId) != _objectDetections.end())
	    _wExtractor->getFalsePositives(negWindows, _objectDetections[imageId], 
					  _threshold, _numNeg);

	  // (2) if any room still, get other objects if necessary
	  int numLeft = _numNeg - negWindows.size();
	  if (!_bFPonly && _bIncludeOthers && numLeft > 0)
	    _wExtractor->getOtherObjects(negWindows, numLeft);

	  // (3) if any room still, extract random detections
	  numLeft = _numNeg - negWindows.size();
	  if (!_bFPonly && numLeft > 0)
	    // hack for now: will oversample, so that when some are discarded based on being uniform
	    // or not interesting enough, the correct number will still remain with high probability
	    _wExtractor->getRandomNegatives(negWindows, numLeft * 3, _negAspectRatio, _negHeight);

	}
      }

      if (posWindows.size() > 0)
	_writer->writeClips(image, posWindows, chPatchType, chExt, true);
      if (negWindows.size() > 0)
	_writer->writeClips(image, negWindows, chPatchType, chExt, false, _numNeg);
      
      cvReleaseImage(&image);

    } //end iterating through images
      
    _writer->printCounts();
    _writer->clearCounts();
  } //end iterating through channels

  // will do so only if required
  _writer->writeWindowReferences();
}








// svlClipWriterClass implementation -------------------------------------------------------------------------------------

svlClipWriter::svlClipWriter() :
  _baseDirectory(svlTrainingDatasetBuilder::BASE_DIR), 
  _negSubdirName(svlTrainingDatasetBuilder::NEG_SUBDIR_NAME),
  _createPosDirs(svlTrainingDatasetBuilder::CREATE_DIRS), 
  _posPrefix(svlTrainingDatasetBuilder::POS_PREFIX),
  _negPrefix(svlTrainingDatasetBuilder::NEG_PREFIX), 
  _bIncludeFlipped(svlTrainingDatasetBuilder::INCLUDE_FLIPPED),
  _resizeWidth(svlTrainingDatasetBuilder::RESIZE_WIDTH), 
  _resizeHeight(svlTrainingDatasetBuilder::RESIZE_HEIGHT),
  _useWindowReferences(svlTrainingDatasetBuilder::USE_WIN_REFS),
  _negCount(0)
{
  // do nothing
}

static void savePatch(const string & filename, const IplImage * patch, svlPatchDefinitionType type)
{
  switch(type)
    {
    case SVL_INTENSITY_PATCH:
      cvSaveImage(filename.c_str(), patch);
      break;
    case SVL_DEPTH_PATCH:
      writeMatrixAsIplImage(patch, filename.c_str());
      break;
    default:
      assert(false); //unsupported patch type
    }
}

void svlClipWriter::writeClips(IplImage *image, const svlObject2dFrame &objects,
			       const svlPatchDefinitionType pType, const string &ext,
			       bool positive, int maxNumToWrite)
{
  // counts the number of examples written in the current function call
  int written = 0;
  for (unsigned i = 0; i < objects.size() && written < maxNumToWrite; i++) {
    svlObject2d object = objects[i];
    
    cvSetImageROI(image, object.getRect());
    CvRect rect = cvGetImageROI(image);

    // check for uniformity of random extractions from intensity images
    if (pType == SVL_INTENSITY_PATCH &&
	object.name == svlImageWindowExtractor::RANDOM_NEG_NAME && 
	svlImageAlmostUniform(image)) {
      cvResetImageROI(image);
      continue;
    }

    //patchtype specific format for image patch to clip out
    IplImage * clip = NULL;
    switch (pType) {
    case SVL_INTENSITY_PATCH:
      clip = cvCreateImage(cvSize(rect.width, rect.height), IPL_DEPTH_8U, 1);
      break;
    case SVL_DEPTH_PATCH:
      clip = cvCreateImage(cvSize(rect.width, rect.height), IPL_DEPTH_32F, 1);
      break;		   
    default:
      assert(false); //unsupported or invalid patch type
    }

   assert(clip);

   if(!_useWindowReferences) //save some time if we're just spitting out references
     cvCopy(image, clip);
  
   cvResetImageROI(image);
  
   if (_resizeWidth > 0 && _resizeHeight > 0)
     resizeInPlace(&clip, _resizeHeight, _resizeWidth);
  
   // extract current index and update count
   unsigned idx = 0;
   if (positive) {
     _it = _posCounts.find(object.name);
     if (_it == _posCounts.end()) {
       _posCounts[object.name] = 1;

       // must be first time ever we're seeing this object (if already
       // saw in a previous channel, _posCounts would've had an entry,
       // maybe 0)
       if (_createPosDirs) {
	 string dirName = string(_baseDirectory) + string("/") + object.name;
	 svlCreateDirectory(dirName.c_str());
       }                    
     } else {
       idx = _it->second++;
     }
   } else {
     idx = _negCount++;
   }

   // creates the filename
   string baseFilename = _baseDirectory + string("/");
   if (positive) {
     baseFilename += object.name + string("/") + _posPrefix + object.name;
   } else {
     // Note: can use object.name for negatives as well, to know where
     // the clip came from originally (it could've been a false positive
     // detection, or random, or a groundtruth labeling of an object
     // we're not currently considering)
     baseFilename += _negSubdirName + string("/") + _negPrefix + string("image");
   }

   baseFilename += padString(toString(idx), 5);
   
   string filename = baseFilename + ext;
   
   if(!_useWindowReferences) {
     savePatch(filename, clip, pType);

     if(_bIncludeFlipped) {
       cvFlip(clip, NULL, 1);
       string flippedFilename = baseFilename + "f" + ext;
       savePatch(flippedFilename.c_str(), clip, pType);
     }
   }
   else {
     if (!positive) object.name = _negSubdirName;
     _windowReferences[filename].push_back(object);

     //(ethand): could add a bool to the references and spit out a
      //second reference here, for now flipping does nothing when
      //-useWindowReferences is enabled
   }
   
   cvReleaseImage(&clip);
   written++;
  }
}

void svlClipWriter::writeWindowReferences()
{
  if (_useWindowReferences) {
    string outputFilename = string(_baseDirectory) + string("/references.xml");
    writeObject2dSequence(outputFilename.c_str(), _windowReferences);
  }
}

void svlClipWriter::clearCounts() 
{
  _negCount = 0;
  for (_it = _posCounts.begin(); _it != _posCounts.end(); _it++) {
    _it->second = 0;
  }
}

void svlClipWriter::printCounts()
{
  SVL_LOG(SVL_LOG_MESSAGE, "Generated dataset with:");
  for (_it = _posCounts.begin(); _it != _posCounts.end(); ++_it)  {
    SVL_LOG(SVL_LOG_MESSAGE, _it->second << " instances of " << _it->first);
  }
  
  SVL_LOG(SVL_LOG_MESSAGE, _negCount << " negative examples");
}






// Configuration class ----------------------------------------------------------------------------------

class svlTrainingDatasetBuilderConfig : public svlConfigurableModule {
public:
  svlTrainingDatasetBuilderConfig() : svlConfigurableModule("svlVision.svlTrainingDatasetBuilder") {}

  void readConfiguration(XMLNode& node) {
    // process everything except the objects
    ((svlConfigurableModule *)this)->readConfiguration(node);

    // parse <object name=""/> (append name to existing objects)
    for (int i = 0; i < node.nChildNode("object"); i++) {
        XMLNode child = node.getChildNode("object", i);
        const char *object = child.getAttribute("name");
        SVL_ASSERT(object != NULL);
	svlTrainingDatasetBuilder::OBJECTS.insert(string(object));
    }
  }

  void setConfiguration(const char *name, const char *value) {
    // just for simplicity, used only if the name corresponds to a boolean
    // variable
    bool valueTrue = (!strcasecmp(value, "TRUE") || !strcmp(value, "1"));

    // related to all extractions
    if (!strcmp(name, "objects")) {
      svlTrainingDatasetBuilder::OBJECTS.clear(); // clear so command line can override config file
      string value_copy(value);
      const char *ptr = strtok((char *)value_copy.c_str(), " ");
      while (ptr != NULL) {
	svlTrainingDatasetBuilder::OBJECTS.insert(string(ptr));
	ptr = strtok(NULL, " ");
      }
    } else if (!strcmp(name, "allWindows")) {
      svlTrainingDatasetBuilder::ALL_WINDOWS = valueTrue;
      SVL_ASSERT_MSG(!svlTrainingDatasetBuilder::POS_BEST_WINDOW,
		     "allWindows incompatible with posBestWindow");

      //related to negatives
    } else if (!strcmp(name, "skipNeg")) {
      svlTrainingDatasetBuilder::SKIP_NEG = valueTrue;
    } else if (!strcmp(name, "falsePositivesOnly")) {
      svlTrainingDatasetBuilder::FPOS_ONLY = valueTrue;
    } else if (!strcmp(name, "detectionsFilename")) {
      svlTrainingDatasetBuilder::DETECTIONS_FILE = string(value);
    } else if (!strcmp(name, "threshold")) {
      svlTrainingDatasetBuilder::THRESHOLD = atof(value);
    } else if (!strcmp(name, "includeOtherObjects")) {
      svlTrainingDatasetBuilder::INCLUDE_OTHER_OBJS = valueTrue;
    } else if (!strcmp(name, "numNegs")) {
      svlTrainingDatasetBuilder::NUM_NEG = atoi(value);
    } else if (!strcmp(name, "negHeight")) {
      svlTrainingDatasetBuilder::NEG_HEIGHT = atoi(value);
    } else if (!strcmp(name, "negAspectRatio")) {
      svlTrainingDatasetBuilder::NEG_ASPECT_RATIO = atof(value);

      // related to positives
    } else if (!strcmp(name, "skipPos")) {
      svlTrainingDatasetBuilder::SKIP_POS = valueTrue;
    } else if (!strcmp(name, "posBestWindow")) {
      svlTrainingDatasetBuilder::POS_BEST_WINDOW = valueTrue;
      SVL_ASSERT_MSG(!svlTrainingDatasetBuilder::ALL_WINDOWS,
		     "allWindows incompatible with posBestWindow");

      // related to writer
    } else if (!strcmp(name, "baseDir")) {
      svlTrainingDatasetBuilder::BASE_DIR = string(value);
    } else if (!strcmp(name, "createDirs")) {
      svlTrainingDatasetBuilder::CREATE_DIRS = valueTrue;
    } else if (!strcmp(name, "negSubdirName")) {
      svlTrainingDatasetBuilder::NEG_SUBDIR_NAME = string(value);
    } else if (!strcmp(name, "posPrefix")) {
      svlTrainingDatasetBuilder::POS_PREFIX = string(value);
    } else if (!strcmp(name, "negPrefix")) {
      svlTrainingDatasetBuilder::NEG_PREFIX = string(value);
    } else if (!strcmp(name, "includeFlipped")) {
      svlTrainingDatasetBuilder::INCLUDE_FLIPPED = valueTrue;
    } else if (!strcmp(name, "resizeWidth")) {
      svlTrainingDatasetBuilder::RESIZE_WIDTH = atoi(value);
    } else if (!strcmp(name, "resizeHeight")) {
      svlTrainingDatasetBuilder::RESIZE_HEIGHT = atoi(value);
    } else if (!strcmp(name, "useWinRefs")) {
      svlTrainingDatasetBuilder::USE_WIN_REFS = valueTrue;
    } else {
      SVL_LOG(SVL_LOG_FATAL, "unrecognized configuration option for " << this->name());
    }
  }
};

static svlTrainingDatasetBuilderConfig gTrainingDatasetBuilderConfig;

